#include <gdk-pixbuf/gdk-pixbuf.h>

#include "ahg-item.h"
#include "ahoghill-media-art.h"
#include "ahoghill-thumbnailer.h"

enum {
    PROP_0,
    PROP_ITEM,
};

enum {
    CLICKED,
    LAST_SIGNAL,
};

struct _AhoghillMediaArtPrivate {
    ClutterActor *art;
    AhgItem *item;
    double rotation;
    ClutterActor *play_texture; /* This is just a clone of play_texture below */

    guint32 update_id;
};

#define GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), AHOGHILL_TYPE_MEDIA_ART, AhoghillMediaArtPrivate))
G_DEFINE_TYPE (AhoghillMediaArt, ahoghill_media_art, MX_TYPE_WIDGET);

static guint32 signals[LAST_SIGNAL] = {0};

static ClutterActor *play_texture = NULL;

static GdkPixbuf *generic_album = NULL;
static GdkPixbuf *generic_image = NULL;
static GdkPixbuf *generic_video = NULL;

static void
ahoghill_media_art_finalize (GObject *object)
{
    g_signal_handlers_destroy (object);
    G_OBJECT_CLASS (ahoghill_media_art_parent_class)->finalize (object);
}

static void
ahoghill_media_art_dispose (GObject *object)
{
    G_OBJECT_CLASS (ahoghill_media_art_parent_class)->dispose (object);
}

static void
clear_texture (AhoghillMediaArt *art)
{
    AhoghillMediaArtPrivate *priv = art->priv;
    GError *error = NULL;
    guint32 data = 0;

    clutter_texture_set_from_rgb_data (CLUTTER_TEXTURE (priv->art),
                                       (const guchar *) &data, TRUE,
                                       1, 1, 1, 4, 0, &error);
    if (error) {
        g_warning ("Error clearing texture: %s", error->message);
        g_error_free (error);
    }
}

static void
set_texture_from_pixbuf (ClutterActor *texture,
                         GdkPixbuf    *thumbnail)
{
    GError *error = NULL;
    gboolean has_alpha;

    has_alpha = gdk_pixbuf_get_has_alpha (thumbnail);
    clutter_texture_set_from_rgb_data (CLUTTER_TEXTURE (texture),
                                       gdk_pixbuf_get_pixels (thumbnail),
                                       has_alpha,
                                       gdk_pixbuf_get_width (thumbnail),
                                       gdk_pixbuf_get_height (thumbnail),
                                       gdk_pixbuf_get_rowstride (thumbnail),
                                       has_alpha ? 4 : 3, 0,
                                       &error);
    if (error != NULL) {
        g_warning ("Error setting %s", error->message);
        g_error_free (error);
    }
}

static void
use_default_texture (AhoghillMediaArt *art,
                     AhgItem          *item)
{
    AhoghillMediaArtPrivate *priv = art->priv;
    GdkPixbuf *thumbnail;
    GError *error = NULL;

    switch (ahg_item_get_item_type (item)) {
    case AHG_ITEM_TYPE_AUDIO:
        if (G_UNLIKELY (generic_album == NULL)) {
            generic_album = gdk_pixbuf_new_from_file
                (THEMEDIR "/hrn-generic-album.png",
                 &error);

            if (generic_album == NULL) {
                g_warning ("Error loading Generic album image: %s",
                           error->message);
                g_error_free (error);
                clear_texture (art);
                return;
            }
        }

        thumbnail = generic_album;
        break;

    case AHG_ITEM_TYPE_IMAGE:
        if (G_UNLIKELY (generic_image == NULL)) {
            generic_image = gdk_pixbuf_new_from_file
                (THEMEDIR "/hrn-generic-images.png",
                 &error);

            if (generic_image == NULL) {
                g_warning ("Error loading Generic image image: %s",
                           error->message);
                g_error_free (error);
                clear_texture (art);
                return;
            }
        }

        thumbnail = generic_image;
        break;

    case AHG_ITEM_TYPE_VIDEO:
        if (G_UNLIKELY (generic_video == NULL)) {
            generic_video = gdk_pixbuf_new_from_file
                (THEMEDIR "/hrn-generic-video.png",
                 &error);

            if (generic_video == NULL) {
                g_warning ("Error loading Generic video image: %s",
                           error->message);
                g_error_free (error);
                clear_texture (art);
                return;
            }
        }

        thumbnail = generic_video;
        break;

    default:
        clear_texture (art);
        return;
    }

    set_texture_from_pixbuf (priv->art, thumbnail);
}

static void
use_texture (AhoghillMediaArt *art,
             const char       *filename)
{
    AhoghillMediaArtPrivate *priv = art->priv;
    GdkPixbuf *pixbuf;
    GError *error = NULL;

    pixbuf = gdk_pixbuf_new_from_file (filename, &error);
    if (error != NULL) {
        g_warning ("Error loading %s: %s", filename, error->message);
        g_error_free (error);

        use_default_texture (art, priv->item);
        return;
    }

    set_texture_from_pixbuf (priv->art, pixbuf);
    g_object_unref (pixbuf);
}

static gboolean
strip_find_next_block (const char    *original,
                       const gunichar open_char,
                       const gunichar close_char,
                       int           *open_pos,
                       int           *close_pos)
{
    const char *p1, *p2;

    if (open_pos) {
        *open_pos = -1;
    }

    if (close_pos) {
        *close_pos = -1;
    }

    p1 = g_utf8_strchr (original, -1, open_char);
    if (p1) {
        if (open_pos) {
            *open_pos = p1 - original;
        }

        p2 = g_utf8_strchr (g_utf8_next_char (p1), -1, close_char);
        if (p2) {
            if (close_pos) {
                *close_pos = p2 - original;
            }

            return TRUE;
        }
    }

    return FALSE;
}

char *
albumart_strip_invalid_entities (const char *original)
{
    GString *str_no_blocks;
    char **strv;
    char *str;
    gboolean blocks_done = FALSE;
    const char *p;
    const char *invalid_chars = "()[]<>{}_!@#$^&*+=|\\/\"'?~";
    const char *invalid_chars_delimiter = "*";
    const char *convert_chars = "\t";
    const char *convert_chars_delimiter = " ";
    const gunichar blocks[5][2] = {
        { '(', ')' },
        { '{', '}' },
        { '[', ']' },
        { '<', '>' },
        {  0,   0  }
    };

    str_no_blocks = g_string_new ("");

    p = original;

    while (!blocks_done) {
        gint pos1, pos2, i;

        pos1 = -1;
        pos2 = -1;

        for (i = 0; blocks[i][0] != 0; i++) {
            gint start, end;

            /* Go through blocks, find the earliest block we can */
            if (strip_find_next_block (p, blocks[i][0], blocks[i][1], &start, &end)) {
                if (pos1 == -1 || start < pos1) {
                    pos1 = start;
                    pos2 = end;
                }
            }
        }

        /* If either are -1 we didn't find any */
        if (pos1 == -1) {
            /* This means no blocks were found */
            g_string_append (str_no_blocks, p);
            blocks_done = TRUE;
        } else {
            /* Append the test BEFORE the block */
            if (pos1 > 0) {
                g_string_append_len (str_no_blocks, p, pos1);
            }

            p = g_utf8_next_char (p + pos2);

            /* Do same again for position AFTER block */
            if (*p == '\0') {
                blocks_done = TRUE;
            }
        }
    }

    str = g_string_free (str_no_blocks, FALSE);

    /* Now strip invalid chars */
    g_strdelimit (str, invalid_chars, *invalid_chars_delimiter);
    strv = g_strsplit (str, invalid_chars_delimiter, -1);
    g_free (str);
    str = g_strjoinv (NULL, strv);
    g_strfreev (strv);

    /* Now convert chars */
    g_strdelimit (str, convert_chars, *convert_chars_delimiter);
    strv = g_strsplit (str, convert_chars_delimiter, -1);
    g_free (str);
    str = g_strjoinv (convert_chars_delimiter, strv);
    g_strfreev (strv);

    /* Now remove double spaces */
    strv = g_strsplit (str, "  ", -1);
    g_free (str);
    str = g_strjoinv (" ", strv);
    g_strfreev (strv);

    /* Now strip leading/trailing white space */
    g_strstrip (str);

    return str;
}

static char *
checksum_string (const char *string)
{
    GChecksum *checksum;
    char *sum;

    checksum = g_checksum_new (G_CHECKSUM_MD5);
    g_checksum_update (checksum, (const guchar *) string, -1);
    sum = g_strdup (g_checksum_get_string (checksum));
    g_checksum_free (checksum);

    return sum;
}

static char *
get_mediaart_uri (const char *first,
                  const char *second,
                  const char *third)
{
    char *artist_esc, *album_esc, *artist_d, *album_d, *b, *c;
    char *dir, *uri;

    artist_esc = second ? albumart_strip_invalid_entities (second) : g_strdup (" ");
    album_esc = third ? albumart_strip_invalid_entities (third) : g_strdup (" ");

    artist_d = g_utf8_strdown (artist_esc, -1);
    album_d = g_utf8_strdown (album_esc, -1);
    g_free (artist_esc);
    g_free (album_esc);

    dir = g_build_filename (g_get_user_cache_dir (), "media-art", NULL);

    b = checksum_string (artist_d);
    c = checksum_string (album_d);
    g_free (artist_d);
    g_free (album_d);

    uri = g_strdup_printf ("file://%s/%s-%s-%s.jpeg", dir, first, b, c);
    g_free (dir);
    g_free (b);
    g_free (c);

    return uri;
}

static char *
get_audio_mediaart_uri (AhgItem *item)
{
    return get_mediaart_uri ("album",
                             ahg_item_get_metadata (item, AHG_ITEM_METADATA_ARTIST),
                             ahg_item_get_metadata (item, AHG_ITEM_METADATA_ALBUM));
}

static char *
make_thumbnailable_uri (AhgItem  *item,
                        char    **mimetype)
{
    switch (ahg_item_get_item_type (item)) {
    case AHG_ITEM_TYPE_IMAGE:
        *mimetype = ahg_item_get_mimetype (item);
        return g_strdup (ahg_item_get_uri (item));

    case AHG_ITEM_TYPE_AUDIO:
        *mimetype = "image/jpeg";
        return get_audio_mediaart_uri (item);

    case AHG_ITEM_TYPE_VIDEO:
        *mimetype = ahg_item_get_mimetype (item);
        return g_strdup (ahg_item_get_uri (item));

    default:
        return NULL;
    }
}

static char *
make_thumbnail_uri (const char *thumbnailable)
{
    GChecksum *checksum;
    char *retval;

    checksum = g_checksum_new (G_CHECKSUM_MD5);
    g_checksum_update (checksum, (const guchar *) thumbnailable, -1);
    retval = g_strdup_printf ("file://%s/.thumbnails/large/%s.png",
                              g_get_home_dir (),
                              g_checksum_get_string (checksum));
    g_checksum_free (checksum);

    return retval;
}

static void
thumbnail_completed (AhoghillThumbnailer *thumbnailer,
                     const char          *request_uri,
                     int                  code,
                     const char          *error,
                     AhoghillMediaArt    *art)
{
    AhoghillMediaArtPrivate *priv = art->priv;
    char *thumbnail_uri;

    if (error != NULL) {
        g_warning ("Error getting thumbnail %s: %s (%d)",
                   request_uri, error, code);
        use_default_texture (art, priv->item);
        return;
    }

    thumbnail_uri = make_thumbnail_uri (request_uri);
    use_texture (art, thumbnail_uri + 7);
    g_free (thumbnail_uri);
}

static gboolean
set_thumbnail (AhoghillMediaArt *art,
               AhgItem          *item)
{
    AhoghillThumbnailer *thumbnailer;
    char *thumbnailable, *thumbnail, *mimetype;

    thumbnailable = make_thumbnailable_uri (item, &mimetype);
    if (thumbnailable == NULL) {
        return FALSE;
    }

    if (g_file_test (thumbnailable + 7, G_FILE_TEST_EXISTS) == FALSE) {
        use_default_texture (art, item);
        g_free (thumbnailable);
        return TRUE;
    }

    thumbnail = make_thumbnail_uri (thumbnailable);
    if (g_file_test (thumbnail + 7, G_FILE_TEST_EXISTS)) {
        use_texture (art, thumbnail + 7);
        g_free (thumbnailable);
        g_free (thumbnail);
        return TRUE;
    }

    g_free (thumbnail);

    /* The thumbnail doesn't exist, so we need to ask the thumbnailer
       daemon for it */
    thumbnailer = ahoghill_thumbnailer_get_default ();
    ahoghill_thumbnailer_request_thumbnail (thumbnailer, thumbnailable,
                                            mimetype,
                                            thumbnail_completed, art);

    return TRUE;
}

static void
item_updated_cb (AhgItem          *item,
                 AhoghillMediaArt *art)
{
    AhoghillMediaArtPrivate *priv = art->priv;

    if (set_thumbnail (art, item)) {
        /* We've got a thumbnail now, we can remove the signal */
        g_signal_handler_disconnect (item, priv->update_id);
        priv->update_id = 0;
    }
}

static void
ahoghill_media_art_set_property (GObject      *object,
                                 guint         prop_id,
                                 const GValue *value,
                                 GParamSpec   *pspec)
{
    AhoghillMediaArt *self = (AhoghillMediaArt *) object;
    AhoghillMediaArtPrivate *priv = self->priv;
    AhgItem *item;

    switch (prop_id) {

    case PROP_ITEM:
        item = g_value_get_object (value);
        if (priv->item == item) {
            break;
        }

        if (priv->item) {
            if (priv->update_id > 0) {
                g_signal_handler_disconnect (priv->item, priv->update_id);
                priv->update_id = 0;
            }

            g_object_unref (priv->item);
            priv->item = NULL;
        }

        if (item) {
            priv->item = g_object_ref (item);
            if (set_thumbnail (self, priv->item) == FALSE) {
                /* Need to hook up to the change to get a thumbnail */
                priv->update_id = g_signal_connect
                    (priv->item, "changed",
                     G_CALLBACK (item_updated_cb), self);
                clear_texture (self);
            }
        } else {
            clear_texture (self);
        }
        break;

    default:
        break;
    }
}

static void
ahoghill_media_art_get_property (GObject    *object,
                                 guint       prop_id,
                                 GValue     *value,
                                 GParamSpec *pspec)
{
    switch (prop_id) {

    default:
        break;
    }
}

static void
ahoghill_media_art_allocate (ClutterActor          *actor,
                             const ClutterActorBox *box,
                             ClutterAllocationFlags flags)
{
    AhoghillMediaArtPrivate *priv = ((AhoghillMediaArt *) actor)->priv;
    ClutterActorClass *parent_class;
    ClutterActorBox child_box, play_box;
    MxPadding padding = { 0, };
    gfloat w, h, x_off, y_off;

    mx_widget_get_padding (MX_WIDGET (actor), &padding);

    parent_class = CLUTTER_ACTOR_CLASS (ahoghill_media_art_parent_class);
    parent_class->allocate (actor, box, flags);

    child_box.x1 = 10;
    child_box.y1 = 7;
    child_box.x2 = box->x2 - box->x1 - 12;
    child_box.y2 = box->y2 - box->y1 - 12;

    clutter_actor_allocate (priv->art, &child_box, flags);

    w = clutter_actor_get_width (priv->play_texture);
    h = clutter_actor_get_height (priv->play_texture);

    x_off = ((child_box.x2 - child_box.x1) - w) / 2;
    y_off = ((child_box.y2 - child_box.y1) - h) / 2;

    play_box.x1 = child_box.x1 + x_off;
    play_box.y1 = child_box.y1 + y_off;
    play_box.x2 = play_box.x1 + w;
    play_box.y2 = play_box.y1 + h;

    clutter_actor_allocate (priv->play_texture, &play_box, flags);
}

/* Taken from PengeMagicTexture */
static void
paint_art (ClutterActor *actor)
{
    ClutterActorBox box;
    CoglHandle *tex;
    int bw, bh;
    int aw, ah;
    float v;
    float tx1, tx2, ty1, ty2;
    ClutterColor col = { 0xff, 0xff, 0xff, 0xff };

    clutter_actor_get_allocation_box (actor, &box);
    tex = clutter_texture_get_cogl_texture ((ClutterTexture *) actor);

    bw = cogl_texture_get_width (tex); /* base texture width */
    bh = cogl_texture_get_height (tex); /* base texture height */

    aw = (int) (box.x2 - box.x1); /* allocation width */
    ah = (int) (box.y2 - box.y1); /* allocation height */

    /* no comment */
    if ((float)bw/bh < (float)aw/ah) {
        /* fit width */
        v = (((float)ah * bw) / ((float)aw * bh)) / 2;
        tx1 = 0;
        tx2 = 1;
        ty1 = (0.5 - v);
        ty2 = (0.5 + v);
    } else {
        /* fit height */
        v = (((float)aw * bh) / ((float)ah * bw)) / 2;
        tx1 = (0.5 - v);
        tx2 = (0.5 + v);
        ty1 = 0;
        ty2 = 1;
    }

    col.alpha = clutter_actor_get_paint_opacity (actor);
    cogl_set_source_color4ub (col.red, col.green, col.blue, col.alpha);
    cogl_rectangle (box.x1, box.y1, box.x2, box.y2);
    cogl_set_source_texture (tex);
    cogl_rectangle_with_texture_coords (box.x1, box.y1,
                                        box.x2, box.y2,
                                        tx1, ty1,
                                        tx2, ty2);
}

static void
ahoghill_media_art_paint (ClutterActor *actor)
{
    AhoghillMediaArtPrivate *priv = ((AhoghillMediaArt *) actor)->priv;

    CLUTTER_ACTOR_CLASS (ahoghill_media_art_parent_class)->paint (actor);

    paint_art (priv->art);

    if (CLUTTER_ACTOR_IS_MAPPED (priv->play_texture)) {
        clutter_actor_paint (priv->play_texture);
    }
}

static void
ahoghill_media_art_map (ClutterActor *actor)
{
    AhoghillMediaArt *art = (AhoghillMediaArt *) actor;
    AhoghillMediaArtPrivate *priv = art->priv;

    CLUTTER_ACTOR_CLASS (ahoghill_media_art_parent_class)->map (actor);

    if (priv->art) {
        clutter_actor_map (priv->art);
    }

    if (priv->play_texture) {
        clutter_actor_map (priv->play_texture);
    }
}

static void
ahoghill_media_art_unmap (ClutterActor *actor)
{
    AhoghillMediaArt *art = (AhoghillMediaArt *) actor;
    AhoghillMediaArtPrivate *priv = art->priv;

    CLUTTER_ACTOR_CLASS (ahoghill_media_art_parent_class)->unmap (actor);

    if (priv->art) {
        clutter_actor_unmap (priv->art);
    }

    if (priv->play_texture) {
        clutter_actor_unmap (priv->play_texture);
    }
}

static gboolean
ahoghill_media_art_enter (ClutterActor         *actor,
                          ClutterCrossingEvent *event)
{
    AhoghillMediaArt *art = (AhoghillMediaArt *) actor;
    AhoghillMediaArtPrivate *priv = art->priv;

    mx_stylable_set_style_pseudo_class (MX_STYLABLE (actor), "hover");

    clutter_actor_animate (priv->play_texture, CLUTTER_EASE_OUT_EXPO, 2000,
                           "opacity", 0xff,
                           NULL);
    return FALSE;
}

static gboolean
ahoghill_media_art_leave (ClutterActor         *actor,
                          ClutterCrossingEvent *event)
{
    AhoghillMediaArt *art = (AhoghillMediaArt *) actor;
    AhoghillMediaArtPrivate *priv = art->priv;

    mx_stylable_set_style_pseudo_class (MX_STYLABLE (actor), NULL);

    clutter_actor_animate (priv->play_texture, CLUTTER_EASE_OUT_EXPO, 2000,
                           "opacity", 0x00,
                           NULL);
    return FALSE;
}

static gboolean
ahoghill_media_art_button_release_event (ClutterActor       *actor,
                                         ClutterButtonEvent *event)
{
    if (event->button == 1) {
        g_signal_emit (actor, signals[CLICKED], 0);
        return TRUE;
    }

    return FALSE;
}

static void
ahoghill_media_art_class_init (AhoghillMediaArtClass *klass)
{
    GObjectClass *o_class = (GObjectClass *)klass;
    ClutterActorClass *a_class = (ClutterActorClass *) klass;

    o_class->dispose = ahoghill_media_art_dispose;
    o_class->finalize = ahoghill_media_art_finalize;
    o_class->set_property = ahoghill_media_art_set_property;
    o_class->get_property = ahoghill_media_art_get_property;

    a_class->paint = ahoghill_media_art_paint;
    a_class->allocate = ahoghill_media_art_allocate;
    a_class->map = ahoghill_media_art_map;
    a_class->unmap = ahoghill_media_art_unmap;
    a_class->enter_event = ahoghill_media_art_enter;
    a_class->leave_event = ahoghill_media_art_leave;
    a_class->button_release_event = ahoghill_media_art_button_release_event;

    g_type_class_add_private (klass, sizeof (AhoghillMediaArtPrivate));
    g_object_class_install_property (o_class, PROP_ITEM,
                                     g_param_spec_object ("item", "", "",
                                                          AHG_TYPE_ITEM,
                                                          G_PARAM_WRITABLE |
                                                          G_PARAM_STATIC_STRINGS));

    signals[CLICKED] = g_signal_new ("clicked",
                                     G_TYPE_FROM_CLASS (o_class),
                                     G_SIGNAL_RUN_FIRST |
                                     G_SIGNAL_NO_RECURSE, 0, NULL, NULL,
                                     g_cclosure_marshal_VOID__VOID,
                                     G_TYPE_NONE, 0);
}

static void
ahoghill_media_art_init (AhoghillMediaArt *self)
{
    AhoghillMediaArtPrivate *priv = GET_PRIVATE (self);
    MxPadding padding = { 0, };

    self->priv = priv;

    mx_widget_get_padding (MX_WIDGET (self), &padding);

    clutter_actor_set_reactive ((ClutterActor *) self, TRUE);

    priv->art = clutter_texture_new ();
    clutter_actor_set_parent ((ClutterActor *) priv->art,
                              (ClutterActor *) self);
    clutter_actor_show ((ClutterActor *) priv->art);

    if (G_UNLIKELY (play_texture == NULL)) {
        GError *error = NULL;

        play_texture = clutter_texture_new_from_file (THEMEDIR "/play_hover.png",
                                                      &error);
        if (play_texture == NULL) {
            g_warning ("Error loading play texture: %s", error->message);
            g_error_free (error);
        } else {
            /* Cloned actors need to be hidden and parented */
            clutter_container_add_actor
                ((ClutterContainer *) clutter_stage_get_default (),
                 play_texture);
            clutter_actor_hide (play_texture);
        }
    }

    if (play_texture) {
        priv->play_texture = clutter_clone_new (play_texture);
        clutter_actor_set_parent (priv->play_texture,
                                  (ClutterActor *) self);
        clutter_actor_set_opacity (priv->play_texture, 0x00);
    }
}

