#include <glib/gi18n.h>

#include <gdk-pixbuf/gdk-pixbuf.h>

#include <mx/mx.h>

#include "ahg-item.h"
#include "ahoghill-media-art.h"
#include "ahoghill-media-tile.h"

enum {
    PROP_0,
    PROP_ITEM,
};

enum {
    CLICKED,
    LAST_SIGNAL
};

struct _AhoghillMediaTilePrivate {
    GdkPixbuf *thumb;
    AhgItem *item;
    guint update_id;

    MxWidget *icon;
    MxWidget *artist;
    MxWidget *title;
};

#define GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), AHOGHILL_TYPE_MEDIA_TILE, AhoghillMediaTilePrivate))
G_DEFINE_TYPE (AhoghillMediaTile, ahoghill_media_tile, MX_TYPE_TABLE);
static guint32 signals[LAST_SIGNAL] = {0, };

#define SPACING 5

#define THUMBNAIL_WIDTH 120
#define THUMBNAIL_HEIGHT 117

static void
ahoghill_media_tile_finalize (GObject *object)
{
    G_OBJECT_CLASS (ahoghill_media_tile_parent_class)->finalize (object);
}

static void
ahoghill_media_tile_dispose (GObject *object)
{
    AhoghillMediaTile *self = (AhoghillMediaTile *) object;
    AhoghillMediaTilePrivate *priv = self->priv;

    if (priv->thumb) {
        g_object_unref (priv->thumb);
        priv->thumb = NULL;
    }

    if (priv->item) {
        g_object_unref (priv->item);
        priv->item = NULL;
    }

    G_OBJECT_CLASS (ahoghill_media_tile_parent_class)->dispose (object);
}

static void
update_tile (AhoghillMediaTile *self)
{
    AhoghillMediaTilePrivate *priv = self->priv;
    GPtrArray *artists;
    const char *title = NULL, *artist = NULL, *width = NULL, *height = NULL;
    const char *season, *episode, *year;
    char *primary = NULL, *secondary = NULL;

    switch (ahg_item_get_item_type (priv->item)) {
    case AHG_ITEM_TYPE_AUDIO:
        title = ahg_item_get_metadata (priv->item, AHG_ITEM_METADATA_TITLE);
        artist = ahg_item_get_metadata (priv->item, AHG_ITEM_METADATA_ARTIST);
        break;

    case AHG_ITEM_TYPE_IMAGE:
        title = ahg_item_get_metadata (priv->item, AHG_ITEM_METADATA_TITLE);
        width = ahg_item_get_metadata (priv->item, AHG_ITEM_METADATA_WIDTH);
        height = ahg_item_get_metadata (priv->item, AHG_ITEM_METADATA_HEIGHT);
        artist = g_strdup_printf ("%sx%s", width, height);
        break;

    case AHG_ITEM_TYPE_VIDEO:
        if (ahg_item_get_metadata (priv->item, AHG_ITEM_METADATA_SERIES_NAME)) {
            title = ahg_item_get_metadata (priv->item,
                                           AHG_ITEM_METADATA_SERIES_NAME);
        } else {
            title = ahg_item_get_metadata (priv->item, AHG_ITEM_METADATA_TITLE);
        }

        year = ahg_item_get_metadata (priv->item, AHG_ITEM_METADATA_DATE);
        if (year == NULL || *year == '\0') {
            season = ahg_item_get_metadata (priv->item,
                                            AHG_ITEM_METADATA_SEASON);
            episode = ahg_item_get_metadata (priv->item,
                                             AHG_ITEM_METADATA_EPISODE);

            /* FIXME: When out of string freeze again, add support for
               "Season %u" and "Episode %u" */
            if ((season == NULL || *season == '\0') &&
                (episode == NULL || *episode == '\0')) {
                secondary = NULL;
            } else {
                secondary = g_strdup_printf (_("Season %s, Episode %s"),
                                           season, episode);
            }
        } else {
            secondary = g_strdup_printf ("(%s)", year);
        }
        break;

    default:
        break;
    }

    if (title == NULL) {
        primary = g_path_get_basename (ahg_item_get_uri (priv->item));
    }

    mx_label_set_text (MX_LABEL (priv->title), title ? title : primary);
    mx_label_set_text (MX_LABEL (priv->artist), artist ? artist :
                       secondary ? secondary : "");

    g_object_set (priv->icon,
                  "item", priv->item,
                  NULL);

    g_free (primary);
    g_free (secondary);
}

static void
clear_tile (AhoghillMediaTile *tile)
{
    AhoghillMediaTilePrivate *priv = tile->priv;

    mx_label_set_text (MX_LABEL (priv->title), "");
    mx_label_set_text (MX_LABEL (priv->artist), "");

    g_object_set (priv->icon,
                  "item", NULL,
                  NULL);
}

static void
item_updated_cb (AhgItem           *item,
                 AhoghillMediaTile *tile)
{
    update_tile (tile);
}

static void
ahoghill_media_tile_set_property (GObject      *object,
                                  guint         prop_id,
                                  const GValue *value,
                                  GParamSpec   *pspec)
{
    AhoghillMediaTile *self = (AhoghillMediaTile *) object;
    AhoghillMediaTilePrivate *priv = self->priv;

    switch (prop_id) {

    case PROP_ITEM:
        if (priv->item) {
            g_signal_handler_disconnect (priv->item, priv->update_id);
            g_object_unref (priv->item);
        }

        priv->item = g_value_dup_object (value);
        if (priv->item) {
            priv->update_id = g_signal_connect (priv->item, "changed",
                                                G_CALLBACK (item_updated_cb),
                                                self);
            update_tile (self);
        } else {
            clear_tile (self);
        }
        break;

    default:
        break;
    }
}

static void
ahoghill_media_tile_get_property (GObject    *object,
                                  guint       prop_id,
                                  GValue     *value,
                                  GParamSpec *pspec)
{
    switch (prop_id) {

    case PROP_ITEM:
    default:
        break;
    }
}

static gboolean
ahoghill_media_tile_button_release_event (ClutterActor       *actor,
                                          ClutterButtonEvent *event)
{
    if (event->button == 1) {
        g_signal_emit (actor, signals[CLICKED], 0);
        return TRUE;
    }

    return FALSE;
}

static void
ahoghill_media_tile_class_init (AhoghillMediaTileClass *klass)
{
    GObjectClass *o_class = (GObjectClass *) klass;
    ClutterActorClass *a_class = (ClutterActorClass *) klass;

    o_class->dispose = ahoghill_media_tile_dispose;
    o_class->finalize = ahoghill_media_tile_finalize;
    o_class->set_property = ahoghill_media_tile_set_property;
    o_class->get_property = ahoghill_media_tile_get_property;

    a_class->button_release_event = ahoghill_media_tile_button_release_event;

    g_type_class_add_private (klass, sizeof (AhoghillMediaTilePrivate));

    g_object_class_install_property (o_class, PROP_ITEM,
                                     g_param_spec_object ("item", "", "",
                                                          AHG_TYPE_ITEM,
                                                          G_PARAM_WRITABLE |
                                                          G_PARAM_STATIC_STRINGS));
    signals[CLICKED] = g_signal_new ("clicked",
                                     G_TYPE_FROM_CLASS (o_class),
                                     G_SIGNAL_RUN_FIRST |
                                     G_SIGNAL_NO_RECURSE, 0, NULL, NULL,
                                     g_cclosure_marshal_VOID__VOID,
                                     G_TYPE_NONE, 0);
}

static void
art_clicked_cb (AhoghillMediaArt  *art,
                AhoghillMediaTile *tile)
{
    g_signal_emit (tile, signals[CLICKED], 0);
}

static void
ahoghill_media_tile_init (AhoghillMediaTile *self)
{
    AhoghillMediaTilePrivate *priv = GET_PRIVATE (self);

    self->priv = priv;

    clutter_actor_set_size (CLUTTER_ACTOR (self), 120, 151);
    clutter_actor_set_reactive (CLUTTER_ACTOR (self), TRUE);

    priv->icon = g_object_new (AHOGHILL_TYPE_MEDIA_ART, NULL);
    g_signal_connect (priv->icon, "clicked",
                      G_CALLBACK (art_clicked_cb), self);
    clutter_actor_set_size (CLUTTER_ACTOR (priv->icon),
                            THUMBNAIL_WIDTH, THUMBNAIL_HEIGHT);
    clutter_actor_show (CLUTTER_ACTOR (priv->icon));

    mx_table_add_actor (MX_TABLE (self), (ClutterActor *) priv->icon, 0, 0);

    priv->title = mx_label_new ("");
    clutter_actor_set_name (CLUTTER_ACTOR (priv->title),
                            "media-tile-title-label");
    mx_table_add_actor_with_properties (MX_TABLE (self),
                                          (ClutterActor *) priv->title, 1, 0,
                                          "x-expand", FALSE,
                                          "x-fill", FALSE,
                                          "y-expand", FALSE,
                                          "y-fill", FALSE,
                                          "x-align", 0.0,
                                          NULL);

    priv->artist = mx_label_new ("");
    clutter_actor_set_name (CLUTTER_ACTOR (priv->artist),
                            "media-tile-artist-label");
    mx_table_add_actor_with_properties (MX_TABLE (self),
                                          (ClutterActor *) priv->artist, 2, 0,
                                          "x-expand", FALSE,
                                          "x-fill", FALSE,
                                          "y-expand", FALSE,
                                          "y-fill", FALSE,
                                          "x-align", 0.0,
                                          NULL);
}
